@@ -30,24 +30,63 @@ analysis of high throughput sequencing data, but if you have useful
 functions for a different domain, please contribute them.
 """
 
-from contextlib import contextmanager
 import pickle
-import pylab
 import re
 import sys
 import os
-from bein import *
-import pysam
 import threading
+from contextlib import contextmanager
+
+from bein import *
 
 # Basic utilities
 
+# 'cat' is used only as an example; it is useless in Bein.
+# @program
+# def _cat(input_file):
+#     return {'arguments': ['cat',input_file],
+#             'return_value': None}
+
+# def cat(ex, input_file, filename=None):
+#     if filename == None:
+#         filename = unique_filename_in()
+#     _cat(ex, input_file, stdout=filename)
+#     return filename
+
+# def _cat_nonblocking(ex, input_file, filename=None):
+#     if filename == None:
+#         filename = unique_filename_in()
+#     f = _cat.nonblocking(ex, input_file, stdout=filename)
+#     class Future(object):
+#         def __init__(self, f):
+#             self.future = f
+#         def wait(self):
+#             self.future.wait()
+#             return filename
+#     return Future(f)
+
+# cat.nonblocking = _cat_nonblocking
+
 
 def pause():
     """Pause until the user hits Return."""
     sys.stdin.readline()
     return None
 
+def first_n_lines(input_file, n, output_file = None):
+    """Writes the first *n* lines of *input_file* to another file.
+
+    If *output_file* is ``None``, then the output is written to a randomly named file.
+    """
+    if output_file == None:
+        output_file = unique_filename_in()
+    with open(input_file, 'r') as inf:
+        with open(output_file, 'w') as outf:
+            for i in xrange(n):
+                l = inf.readline()
+                outf.write(l)
+    return output_file
+            
 
 @program
 def touch(filename = None):
@@ -218,7 +257,8 @@ def bowtie_build(files, index = None):
             'return_value': index}
 
 
-def parallel_bowtie(ex, index, reads, n_lines = 1000000, bowtie_args="-Sra", add_nh_flags=False):
+def parallel_bowtie(ex, index, reads, n_lines = 1000000, 
+                    bowtie_args="-Sra", add_nh_flags=False, via='local'):
     """Run bowtie in parallel on pieces of *reads*.
 
     Splits *reads* into chunks *n_lines* long, then runs bowtie with
@@ -230,17 +270,22 @@ def parallel_bowtie(ex, index, reads, n_lines = 1000000, bowtie_args="-Sra", add
     Bowtie does not set the NH flag on its SAM file output.  If the
     *add_nh_flags* argument is ``True``, this function calculates
     and adds the flag before merging the BAM files.
+
+    The *via* argument determines how the jobs will be run.  The
+    default, ``'local'``, runs them on the same machine in separate
+    threads.  ``'lsf'`` submits them via LSF.
     """
     subfiles = split_file(ex, reads, n_lines = n_lines)
-    futures = deepmap(lambda sf: bowtie.nonblocking(ex, index, sf, args = bowtie_args), 
-                      subfiles)
-    samfiles = deepmap(lambda f: f.wait(), futures)
+    futures = [bowtie.nonblocking(ex, index, sf, args=bowtie_args, via=via)
+               for sf in subfiles]
+    samfiles = [f.wait() for f in futures]
     if add_nh_flags:
-        bamfiles = deepmap(add_nh_flag, samfiles)
+        futures = [external_add_nh_flag.nonblocking(ex, sf, via=via) for sf in samfiles]
+        bamfiles = [f.wait() for f in futures]
     else:
-        futures = deepmap(lambda sf: sam_to_bam.nonblocking(ex, sf), samfiles)
-        bamfiles = deepmap(lambda f: f.wait(), futures)
-    return merge_bam.nonblocking(ex, bamfiles).wait()
+        futures = [sam_to_bam.nonblocking(ex, sf, via=via) for sf in samfiles]
+        bamfiles = [f.wait() for f in futures]
+    return merge_bam.nonblocking(ex, bamfiles, via=via).wait()
 
 def deepmap(f, st):
     """Map function *f* over a structure *st*.
@@ -272,17 +317,15 @@ def deepmap(f, st):
         return f(st)
 
 def parallel_bowtie_lsf(ex, index, reads, n_lines = 1000000, bowtie_args="-Sra", add_nh_flags=False):
-    """Identical to parallel_bowtie, but runs programs via LSF."""
-    subfiles = split_file(ex, reads, n_lines = n_lines)
-    futures = deepmap(lambda sf: bowtie.lsf(ex, index, sf, args = bowtie_args), 
-                      subfiles)
-    samfiles = deepmap(lambda f: f.wait(), futures)
-    if add_nh_flags:
-        bamfiles = deepmap(add_nh_flag, samfiles)
-    else:
-        futures = deepmap(lambda sf: sam_to_bam.lsf(ex, sf), samfiles)
-        bamfiles = deepmap(lambda f: f.wait(), futures)
-    return merge_bam.lsf(ex, bamfiles).wait()
+    """Deprecated.  Use parallel_bowtie with ``via="lsf"`` instead."""
+    raise DeprecationWarning("parallel_bowtie_lsf is deprecated.  Use parallel_bowtie with via='lsf' instead")
+    # return parallel_bowtie(ex, index, reads, n_lines, bowtie_args, add_nh_flags, via='lsf')
+
+@program
+def external_add_nh_flag(samfile):
+    outfile = unique_filename_in()
+    return {'arguments': ['add_nh_flag',samfile,outfile],
+            'return_value': outfile}
 
 ###############
 # BAM/SAM files
@@ -322,7 +365,7 @@ def replace_bam_header(header, bamfile):
 
 @program
 def sort_bam(bamfile):
-    """Sort a BAM file *bamfile*.
+    """Sort a BAM file *bamfile* by chromosome coordinates.
 
     Returns the filename of the newly created, sorted BAM file.
 
@@ -332,6 +375,43 @@ def sort_bam(bamfile):
     return {'arguments': ['samtools','sort',bamfile,filename],
             'return_value': filename + '.bam'}
 
+@program
+def sort_bam_by_read(bamfile):
+    """Sort a BAM file *bamfile* by read names.
+
+    Returns the filename of the newly created, sorted BAM file.
+
+    Equivalent: ``samtools sort -n ...``
+    """
+    filename = unique_filename_in()
+    return {'arguments': ['samtools','sort','-n',bamfile,filename],
+            'return_value': filename + '.bam'}
+
+
+def read_sets(reads):
+    """Groups the alignments in a BAM file by read.
+
+    *reads* should be an iterator over reads, such as the object
+     returned by pysam.Samfile.  The SAM/BAM file must be sorted by
+     read.  ``read_sets`` removes all unmapped reads, and returns an
+     iterator over lists of all AlignedRead objects consisting of the
+     same read.
+    """
+    last_read = None
+    for r in reads:
+        if r.rname == -1 or r.is_unmapped:
+            pass
+        elif r.qname != last_read:
+            if last_read != None:
+                yield accum
+            accum = [r]
+            last_read = r.qname
+        else:
+            accum.append(r)
+    if last_read != None:
+        # We have to check, since if samfile
+        # has no alignments, accum is never defined.
+        yield accum    
 
 @program
 def index_bam(bamfile):
@@ -363,48 +443,36 @@ def merge_bam(files):
         return {'arguments': ['samtools','merge',filename] + files,
                 'return_value': filename}
 
-def split_by_readname(samfile):
-    """Return an iterator over the reads in *samfile* grouped by read name.
 
-    The SAM file produced by bowtie is sorted by read name.  Often we
-    want to work with all of the alignments of a particular read at
-    once.  This function turns the flat list of reads into a list of
-    lists of reads, where each sublist has the same read name.
-    """
-    last_read = None
-    for r in samfile:
-        if r.qname != last_read:
-            if last_read != None:
-                yield accum
-            accum = [r]
-            last_read = r.qname
+try:
+    import pysam
+    def add_nh_flag(samfile, out=None):
+        """Adds NH (Number of Hits) flag to each read alignment in *samfile*.
+        
+        Scans a BAM file ordered by read name, counts the number of
+        alternative alignments reported and writes them to a BAM file
+        with the NH tag added.
+        
+        If *out* is ``None``, a random name is used.
+        """
+        infile = pysam.Samfile(samfile, "r")
+        if out == None:
+            outname = unique_filename_in()
         else:
-            accum.append(r)
-    if last_read != None:
-        # We have to check, since if samfile
-        # has no alignments, accum is never defined.
-        yield accum
-
-def add_nh_flag(samfile):
-    """Adds NH (Number of Hits) flag to each read alignment in *samfile*.
-    
-    Scans a TAM file ordered by read name, counts the number of
-    alternative alignments reported and writes them to a BAM file
-    with the NH tag added.
-    """
-    infile = pysam.Samfile(samfile, "r")
-    outname = unique_filename_in()
-    outfile = pysam.Samfile(outname, "wb", template=infile)
-    for readset in split_by_readname(infile):
-        nh = len(readset)
-        for read in readset:
-            if (read.is_unmapped):
-                nh = 0
-            read.tags = read.tags+[("NH",nh)]
-            outfile.write(read)
-    infile.close()
-    outfile.close()
-    return outname
+            outname = out
+        outfile = pysam.Samfile(outname, "wb", template=infile)
+        for readset in split_by_readname(infile):
+            nh = len(readset)
+            for read in readset:
+                if (read.is_unmapped):
+                    nh = 0
+                read.tags = read.tags+[("NH",nh)]
+                outfile.write(read)
+        infile.close()
+        outfile.close()
+        return outname
+except:
+    print >>sys.stderr, "PySam not found.  Skipping add_nh_flag."
 
 # Adding special file types
 
@@ -423,25 +491,28 @@ def add_pickle(execution, val, description="", alias=None):
     execution.add(filename, description=description, alias=alias)
     return filename
 
-
-@contextmanager
-def add_figure(ex, figure_type='eps', description="", alias=None, figure_size=None):
-    """Create a matplotlib figure and write it to the repository.
-
-    Use this as a with statement, for instance::
-
-        with add_figure(ex, 'eps') as fig:
-            hist(a)
-            xlabel('Random things I found')
-
-    This will plot a histogram of a with the x axis label set, and
-    write the plot to the repository as an EPS file.
-    """
-    f = pylab.figure(figsize=figure_size)
-    yield f
-    filename = unique_filename_in() + '.' + figure_type
-    f.savefig(filename)
-    ex.add(filename, description=description, alias=alias)
+try:
+    import pylab
+    @contextmanager
+    def add_figure(ex, figure_type='eps', description="", alias=None, figure_size=None):
+        """Create a matplotlib figure and write it to the repository.
+    
+        Use this as a with statement, for instance::
+    
+            with add_figure(ex, 'eps') as fig:
+                hist(a)
+                xlabel('Random things I found')
+    
+        This will plot a histogram of a with the x axis label set, and
+        write the plot to the repository as an EPS file.
+        """
+        f = pylab.figure(figsize=figure_size)
+        yield f
+        filename = unique_filename_in() + '.' + figure_type
+        f.savefig(filename)
+        ex.add(filename, description=description, alias=alias)
+except:
+    print >>sys.stderr, "Could not import matplotlib.  Skipping add_figure."
 
 
 def add_and_index_bam(ex, bamfile, description="", alias=None):
@@ -483,6 +554,22 @@ def add_bowtie_index(execution, files, description="", alias=None, index=None):
     execution.add(index + ".rev.2.ebwt", associate_to_filename=index, template='%s.rev.2.ebwt')
     return index
 
+try:
+    import tables as h5
+    @contextmanager
+    def add_hdf5(ex, description='', alias=None):
+        h5filename = unique_filename_in()
+        db = h5.openFile(h5filename, 'w', title=description)
+        try:
+            yield db
+        finally:
+            db.close()
+            ex.add(db, description=description, alias=alias)
+except:
+    print >>sys.stderr, "PyTables not found.  Skipping."
+
+
+
 if __name__ == "__main__":
     import doctest
     doctest.testmod()
